;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#|
	Andy Long
	Mathemagician
	Northern Kentucky University
	Department of Mathematics and Statistics
	Wed Oct 27 10:35:30 EDT 2010

This file creates a graphic of all simple graphs on five vertices. 

It does this by identifying simple graphs with matrices: in this case,
five-by-five matrices, with zeros on the diagonals (no loops), and arcs
indicated by a one in the appropriate column. These matrices are symmetric, so
I exploit that in writing and operating with the matrices. 

The ingredients of the idea are these:
	a. All permutation matrices 5x5;
	b. all subsets of the 10 upper elements of a 5x5 matrices (which could
	   potentially be different from zero, because of an arc);
	c. duality, which permits us to do only graphs with five arcs or fewer.
We create a graph by making a subset of the upper-elements of the 5x5 matrix
ones, and then compare it to all the other (saved, distinct) graphs of the same
number of arcs to see if it's redundant. If not, we include it into the mix of
saved, distinct graphs with the requisite number of arcs. 

We do this for all possibilities up to five arcs, then use duality to get the
rest of the distinct graphs. In particular, my ambition was to create a really
beautiful collection of such graphs -- to arrange them just so in my graphic. I
think that I've succeeded!

Thanks to Sean Bodine for fixing a duality glitch.
|#
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Permutation matrices:
;;
;;  An ugly, kludgy way to create all 120 permutation matrices. I liked this
;;  method because it just involved cutting and pasting a lot!;)
;;  
;;  I'm a little wasteful here, because I consider, for example, the identity
;;  matrix.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def 
 permutations
 (let (permutations)
   (dolist (i1 (iseq 5))
     (let ((new (remove i1 (iseq 5))))
       (dolist (i2 new)
	 (let ((new (remove i2 new)))
	   (dolist (i3 new)
	     (let ((new (remove i3 new)))
	       (dolist (i4 new)
		 (let ((new (remove i4 new)))
		   (dolist (i5 new)
		     (setq
		      permutations
		      (append permutations
			      (list
			       (apply
				#'bind-columns
				(select 
				 (row-list (id 5)) 
				 (list i1 i2 i3 i4 i5)))))))))))))))
   permutations
   )
 )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Here's the conjugation function, AMA'
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun permute-em(M A) (matmult A M (transpose A)) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 
;; Now we create all subsets, possible combinations of the 10 upper-triangular
;; elements of the matrices which may be non-zero: a subset represents the set
;; of non-zero entries. We want to consider all possible combinations of
;; non-zero subsets of these entries.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def subsets (all-subsets (iseq 10)))
(def sizes (mapcar #'size subsets))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; distinct-graph-matrices 
;;
;;  All the different graphs, that is matrices. We initialize the variable
;; with the zero matrix, corresponding to the graph made up of isolated
;; points:
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(def distinct-graph-matrices (list (list (zeros 5 5))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;               Here we go!
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(mapcar 
 ;; The argument i is the number of non-zero elements in the upper-triangular
 ;; part of a matrix (i.e. the number of arcs):
 (lambda(i)
   (let* (
	  ;; we pull off those i-sized subsets,
	  (right-size (select subsets (where sizes i)))
	  ;; then create the matrices:
	  (matrices 
	   (mapcar
	    (lambda(set)
	      (let ((indicators (repeat 0 10)))
		(setf (select indicators set) (repeat 1 (length set)))
		;; I'm creating full matrices, which is wasteful (but easy!):
		(matrix '(5 5)
			(combine
			 0 (select indicators (iseq 4))
			 (elt indicators 0) 0 (select indicators (iseq 4 6))
			 (select indicators (list 1 4)) 0 (select indicators (iseq 7 8))
			 (select indicators (list 2 5 7)) 0 (elt indicators 9)
			 (select indicators (list 3 6 8 9)) 0
			 )
			)
		)
	      )
	    right-size
	    )
	   )
	  ;; Now I'm going to establish a list of all the distinct matrices of
	  ;; this size (and hence distinct graphs). We initialize with the
	  ;; first of the possible matrices, because it may certainly be taken
	  ;; to be a keeper, WLOG...
	  (keepers (list (first matrices)))
	  )
     ;; Now, for the rest of the matrices, we 
     (mapcar
      (lambda(x) 
	(let* (
	       ;; compute all permutations,
	       (all (mapcar (lambda(mat) (permute-em x mat)) permutations))
	       ;; and check to see if any of the permutation is already in the
	       ;; keeper list (we'll check by computing the norms of the
	       ;; differences of the permuted matrices, and all the keepers
	       ;; already established:
	       (norms 
		(combine
		 (mapcar
		  (lambda(x)
		    (mapcar (lambda(y) (norm (- x y))) keepers)
		    )
		  all
		  )
		 )
		)
	       )
	  ;; if any norm is zero, then that matrix (that graph) is already represented, and may be
	  ;; eliminated from consideration. 
	  (if (not (where norms 0))
	      ;; otherwise, we add it to the keeper list: it's a new one!
	      (setq keepers (append keepers (list x)))
	    )
	  )
	)
      (rest matrices)
      )
     ;; we now update our distinct-graph-matrices, with the new group of
     ;; i-sized keepers:
     (setf 
      distinct-graph-matrices
      (append distinct-graph-matrices (list keepers))
      )
     )
   nil
   )
 ;; we only do this for i from 1 to 5: the i=0 case is the zeros matrix, which
 ;; initialized distinct-graph-matrices, and cases above 5 can be handled by
 ;; duality.
 (iseq 1 5)
 )

;; Duality can be used to get the cases above 5: the dual of a graph g is
;; connected where g is unconnected, and unconnected where g is connected).
;; I start with the duals of the graphs with 0-4 arcs:
(def duals
     (mapcar
      (lambda(x)
	(mapcar 
	 (lambda(y) (- (ones 5 5) y (id 5)))
	 x
	 )
	)
      (butlast distinct-graph-matrices)
      )
     )
;; for the five duals, I use half of them as received, and then create the
;; duals of those to get the other half -- this will lead to a more attractive
;; display of graphs. 
#|
(def dual-fives
     (mapcar 
      (lambda(y) (- (ones 5 5) y (id 5)))
      (final  distinct-graph-matrices)
      )
     )
|#
(def dual-fives
     (select
      (final  distinct-graph-matrices)
      (list 2 0 3 4 1 5 ) 
      )
     )

;; I reverse the order, so that when I graph the five-arc graphs, they'll be
;; ordered symmetrically with their duals (they'll be in a stack, and the first
;; three will be the original five-arc graphs, and the next three will be their
;; duals in reverse order -- more attractive that way!
#|
(def fives (append 
	    (select (final distinct-graph-matrices) (iseq 3))
	    (reverse (select dual-fives (iseq 3)))
	    )
     )
|#
(def fives 
     (list
      (elt dual-fives 0)
      (elt dual-fives 1)
      (elt dual-fives 2)
      (- (ones 5 5) (elt dual-fives 2) (id 5))
      (- (ones 5 5) (elt dual-fives 1) (id 5))
      (elt dual-fives 5)
      )
     )

;; Now we put them all back together into a list of 34 graphs:
(def distinct-graph-matrices 
     (smoosh 
      (append
       (butlast distinct-graph-matrices)
       (list fives)
       (reverse duals))))
     
;; We can just check that we have all 34:
(size distinct-graph-matrices)

#|
Now, to some graphics:
|#
;; The points start off as the 5th roots of unity (i.e. I compute them using
;; complex analysis!), then turned into coordinate pairs:
(def
 points  (mapcar 
	  (lambda(i) 
	    (let ((complex 
		   (exp 
		    (+ 
		     (* (complex 0 1) pi/2) 
		     (* (complex 0 1) i (/ 2pi 5))))))
	      (list (realpart complex) (imagpart complex))
	      )
	    )
	  (iseq 5))
 )

;; coords are the indices of the matrix: they'll be used to figure out how to
;; connect up the vertices:
(def coords '((0 1) (0 2) (0 3) (0 4) (1 2) (1 3) (1 4) (2 3) (2 4) (3 4)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 
;;  Okay, this is it: create the graphic for each distinct graph, and position
;; it so that they're arranged to show off the duality -- left to right for the
;; graphs other than the five-arc graphs -- and the five-arc graphs will show
;; off their duality in the vertical sense. 
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(let (
      (coords-y (repeat 0 11))
      morelines
      ;; morepoints is initialized to contain the 5-graph with 0 arcs:
      (morepoints
       (list
	(list
	 (repeat 1 5)
	 (repeat 0 5)
	 (repeat 0 5)
	 (repeat 6 5)
	 (repeat 4 5)
	 (mapcar #'first points)
	 (mapcar #'second points)
	 )
	)
       )
      )
  ;; now: we're going to create all the graphs. Morelines is in charge of the
  ;; arcs, and morepoints is in charge of the vertices.
  (mapcar
   (lambda(x) 
     (let* (
	    (keepers 
	     (select coords (where (combine 
				    (select x 0 (iseq 1 4))
				    (select x 1 (iseq 2 4))
				    (select x 2 (iseq 3 4))
				    (select x 3 (list 4))
				    )
				   1)))
	    (xc (length keepers))
	    (yc (let (
		      (tmp (elt coords-y xc)) 
		      )
		  (incf (elt coords-y xc))
		  tmp))
	    (xshift (* 3 xc))
	    (yshift (* 3 yc))
	    (start-vertices (mapcar #'first keepers))
	    (  end-vertices (mapcar #'second keepers))
	    )
       (setq 
	morelines
	(append 
	 morelines
	 (mapcar (lambda(x y) 
		   (list 0 0 1 
			 (+ xshift
			    (list 
			     (first (elt points x))
			     (first (elt points y))
			     )
			    )
			 (+ yshift
			    (list 
			     (second (elt points x))
			     (second (elt points y))
			     )
			    )
			 'ln)
		   )
		 start-vertices
		 end-vertices
		 )
	 )
	morepoints
	(append 
	 morepoints
	 (list
	  (list
	   (repeat 1 5)
	   (repeat 0 5)
	   (repeat 0 5)
	   (repeat 6 5)
	   (repeat 4 5)
	   (+ xshift (mapcar #'first points) )
	   (+ yshift (mapcar #'second points) )
	   )
	  )
	 )
	)
       )
     )
   (rest distinct-graph-matrices)
   )
  (xgobips
   "graphs.ps"
   0 0 1
   (mapcar #'first points)
   (mapcar #'second points)
   :glyph-type 4
   :morelines
   morelines
   :points
   morepoints
   :xmin -1
   :xmax  31
   :ymin -1
   :ymax 16
   :xtics (* 4 '(0 2 4 6 8))
   :ytics (* 2 '(0 2 4 6 8))
   :title "All simple graphs with 5 vertices -- note the duality"
   )
  )  
